/******************************************************************************
 * $Header: /boot/home/agmsmith/Programming/MailboxFileToBeMail/RCS/MailboxToBeMail.cpp,v 1.6 2002/01/14 20:23:21 agmsmith Exp $
 *
 * This program converts Unix mailbox files (the kind that Pine uses) into mail
 * files for use with BeOS.  It also handles news files from rn and trn, which
 * have messages very similar to mail messages but with a different separator
 * line.  The input files store multiple mail messages in text format separated
 * by "From ..." lines or "Article ..." lines.  The output is a bunch of
 * separate files, each one with one message plus attributes describing that
 * message.  For convenience, all the messages that were from one file are put
 * in a specified directory.  Command line driven.
 *
 * $Log: MailboxToBeMail.cpp,v $
 * Revision 1.6  2002/01/14 20:23:21  agmsmith
 * Include the detection line with the message header, just for completeness.
 *
 * Revision 1.5  2002/01/14 03:21:43  agmsmith
 * Added processing of UseNet articles, and allowed an older date format.
 *
 * Revision 1.4  2002/01/14 00:06:59  agmsmith
 * Don't allow slashes in file names.
 *
 * Revision 1.3  2002/01/13 23:33:26  agmsmith
 * First seemingly working version.
 *
 * Revision 1.2  2002/01/13 17:57:26  agmsmith
 * Can now pick out separate messages from the inbox.
 *
 * Revision 1.1  2002/01/13 15:25:33  agmsmith
 * Initial revision
 */

/* BeOS headers. */

#include <Application.h>
#include <StorageKit.h>
#include <SupportKit.h>

/* Posix headers. */

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <parsedate.h>


/******************************************************************************
 * Globals
 */

char  InputPathName [B_PATH_NAME_LENGTH];
FILE *InputFile;
char  OutputDirectoryPathName [B_PATH_NAME_LENGTH];
  /* Name of the output directory includes an ending slash, or the whole
  string is an empty string if it is using the current directory.  Either
  way, you can prefix the file name with the string to get the file path. */


typedef enum StandardHeaderEnum
{
  STD_HDR_DATE = 0, /* The Date: field. */
  STD_HDR_MIME_VERSION, /* The Mime-Version: field, usually "1.0". */
  STD_HDR_FROM, /* The whole From: field, including quotes and address. */
  STD_HDR_TO, /* All the stuff in the To: field. */
  STD_HDR_CC, /* All the CC: field (originally means carbon copy). */
  STD_HDR_REPLY, /* Things in the reply-to: field. */
  STD_HDR_SUBJECT, /* The Subject: field. */
  STD_HDR_PRIORITY, /* The Priority: field, usually "Normal". */
  STD_HDR_X_PRIORITY, /* The X-Priority: field, usually a number out of 10? */
  STD_HDR_MSMAIL_PRIORITY, /* The X-MAMail-Priority: field, "Normal" etc. */
  STD_HDR_MAX
} StandardHeaderCodes;

const char *g_StandardHeaders [] =
{
  "Date",
  "MIME-Version",
  "From",
  "To",
  "Cc",
  "Reply-To",
  "Subject",
  "Priority",
  "X-Priority",
  "X-MSMail-Priority"
};



/******************************************************************************
 * Global utility function to display an error message and return.  The message
 * part describes the error, and if ErrorNumber is non-zero, gets the string
 * ", error code $X (standard description)." appended to it.  If the message
 * is NULL then it gets defaulted to "Something went wrong".
 */

static void DisplayErrorMessage (
  const char *MessageString = NULL,
  int ErrorNumber = 0,
  const char *TitleString = NULL)
{
  char ErrorBuffer [B_PATH_NAME_LENGTH + 80 /* error message */ + 80];

  if (TitleString == NULL)
    TitleString = "Error Message:";

  if (MessageString == NULL)
  {
    if (ErrorNumber == 0)
      MessageString = "No error, no message, why bother?";
    else
      MessageString = "Something went wrong";
  }

  if (ErrorNumber != 0)
  {
    sprintf (ErrorBuffer, "%s, error code $%X/%d (%s) has occured.",
      MessageString, ErrorNumber, ErrorNumber, strerror (ErrorNumber));
    MessageString = ErrorBuffer;
  }

  fputs (TitleString, stderr);
  fputc ('\n', stderr);
  fputs (MessageString, stderr);
  fputc ('\n', stderr);
}



/******************************************************************************
 * Determine if a line of text is the start of another message.  Pine mailbox
 * files have messages that start with a line that could say something like
 * "From agmsmith@achilles.net Fri Oct 31 21:19:36 EST 1997" or maybe something
 * like "From POPmail Mon Oct 20 21:12:36 1997" or in a more modern format,
 * "From agmsmith@achilles.net Tue Sep 4 09:04:11 2001 -0400".  I generalise it
 * to "From blah Day MMM NN XX:XX:XX TZONE1 YYYY TZONE2".  Blah is an e-mail
 * address you can ignore (just treat it as a word separated by spaces).  Day
 * is a 3 letter day of the week.  MMM is a 3 letter month name.  NN is the two
 * digit day of the week, has a leading space if the day is less than 10.
 * XX:XX:XX is the time, the X's are digits.  TZONE1 is the old style optional
 * time zone of 3 capital letters.  YYYY is the four digit year.  TZONE2 is the
 * optional modern time zone info, a plus or minus sign and 4 digits.  Returns
 * true if the line of text (ended with a NUL byte, no line feed or carriage
 * returns at the end) is the start of a message.
 */

bool IsStartOfMailMessage (char *LineString)
{
  char        *StringPntr;

  /* It starts with "From " */

  if (memcmp ("From ", LineString, 5) != 0)
    return false;
  StringPntr = LineString + 4;
  while (*StringPntr == ' ')
    StringPntr++;

  /* Skip over the e-mail address (or stop at the end of string). */

  while (*StringPntr != ' ' && *StringPntr != 0)
    StringPntr++;
  while (*StringPntr == ' ')
    StringPntr++;

  /* Look for the 3 letter day of the week. */

  if (memcmp (StringPntr, "Mon", 3) != 0 &&
  memcmp (StringPntr, "Tue", 3) != 0 &&
  memcmp (StringPntr, "Wed", 3) != 0 &&
  memcmp (StringPntr, "Thu", 3) != 0 &&
  memcmp (StringPntr, "Fri", 3) != 0 &&
  memcmp (StringPntr, "Sat", 3) != 0 &&
  memcmp (StringPntr, "Sun", 3) != 0)
  {
    printf ("False alarm, not a valid day of the week in \"%s\".\n",
      LineString);
    return false;
  }
  StringPntr += 3;
  while (*StringPntr == ' ')
    StringPntr++;

  /* Look for the 3 letter month code. */

  if (memcmp (StringPntr, "Jan", 3) != 0 &&
  memcmp (StringPntr, "Feb", 3) != 0 &&
  memcmp (StringPntr, "Mar", 3) != 0 &&
  memcmp (StringPntr, "Apr", 3) != 0 &&
  memcmp (StringPntr, "May", 3) != 0 &&
  memcmp (StringPntr, "Jun", 3) != 0 &&
  memcmp (StringPntr, "Jul", 3) != 0 &&
  memcmp (StringPntr, "Aug", 3) != 0 &&
  memcmp (StringPntr, "Sep", 3) != 0 &&
  memcmp (StringPntr, "Oct", 3) != 0 &&
  memcmp (StringPntr, "Nov", 3) != 0 &&
  memcmp (StringPntr, "Dec", 3) != 0)
  {
    printf ("False alarm, not a valid month name in \"%s\".\n",
      LineString);
    return false;
  }
  StringPntr += 3;
  while (*StringPntr == ' ')
    StringPntr++;

  /* Skip the day of the month.  Require at least one digit. */

  if (*StringPntr < '0' || *StringPntr > '9')
  {
    printf ("False alarm, not a valid day of the month number in \"%s\".\n",
      LineString);
    return false;
  }
  while (*StringPntr >= '0' && *StringPntr <= '9')
    StringPntr++;
  while (*StringPntr == ' ')
    StringPntr++;

  /* Check the time.  Look for the sequence
  digit-digit-colon-digit-digit-colon-digit-digit. */

  if (StringPntr[0] < '0' || StringPntr[0] > '9' ||
  StringPntr[1] < '0' || StringPntr[1] > '9' ||
  StringPntr[2] != ':' ||
  StringPntr[3] < '0' || StringPntr[3] > '9' ||
  StringPntr[4] < '0' || StringPntr[4] > '9' ||
  StringPntr[5] != ':' ||
  StringPntr[6] < '0' || StringPntr[6] > '9' ||
  StringPntr[7] < '0' || StringPntr[7] > '9')
  {
    printf ("False alarm, not a valid time value in \"%s\".\n",
      LineString);
    return false;
  }
  StringPntr += 8;
  while (*StringPntr == ' ')
    StringPntr++;

  /* Look for the optional antique 3 capital letter time zone and skip it. */

  if (StringPntr[0] >= 'A' && StringPntr[0] <= 'Z' &&
  StringPntr[1] >= 'A' && StringPntr[1] <= 'Z' &&
  StringPntr[2] >= 'A' && StringPntr[2] <= 'Z')
  {
    StringPntr += 3;
    while (*StringPntr == ' ')
      StringPntr++;
  }

  /* Look for the 4 digit year. */

  if (StringPntr[0] < '0' || StringPntr[0] > '9' ||
  StringPntr[1] < '0' || StringPntr[1] > '9' ||
  StringPntr[2] < '0' || StringPntr[2] > '9' ||
  StringPntr[3] < '0' || StringPntr[3] > '9')
  {
    printf ("False alarm, not a valid 4 digit year in \"%s\".\n",
      LineString);
    return false;
  }
  StringPntr += 4;
  while (*StringPntr == ' ')
    StringPntr++;

  /* Look for the optional modern time zone and skip over it if present. */

  if ((StringPntr[0] == '+' || StringPntr[0] == '-') &&
  StringPntr[1] >= '0' && StringPntr[1] <= '9' &&
  StringPntr[2] >= '0' && StringPntr[2] <= '9' &&
  StringPntr[3] >= '0' && StringPntr[3] <= '9' &&
  StringPntr[4] >= '0' && StringPntr[4] <= '9')
  {
    StringPntr += 5;
    while (*StringPntr == ' ')
      StringPntr++;
  }

  /* Look for end of string. */

  if (*StringPntr != 0)
  {
    printf ("False alarm, extra stuff after the year/time zone in \"%s\".\n",
      LineString);
    return false;
  }

  return true;
}



/******************************************************************************
 * Determine if a line of text is the start of a news article.  TRN and RN news
 * article save files have messages that start with a line that looks like
 * "Article 11721 of rec.games.video.3do:".  Returns true if the line of text
 * (ended with a NUL byte, no line feed or carriage returns at the end) is the
 * start of an article.
 */

bool IsStartOfUsenetArticle (char *LineString)
{
  char        *StringPntr;

  /* It starts with "Article " */

  if (memcmp ("Article ", LineString, 8) != 0)
    return false;
  StringPntr = LineString + 7;
  while (*StringPntr == ' ')
    StringPntr++;

  /* Skip the article number.  Require at least one digit. */

  if (*StringPntr < '0' || *StringPntr > '9')
  {
    printf ("False alarm, not a valid article number in \"%s\".\n",
      LineString);
    return false;
  }
  while (*StringPntr >= '0' && *StringPntr <= '9')
    StringPntr++;
  while (*StringPntr == ' ')
    StringPntr++;

  /* Now it should have "of " */

  if (memcmp ("of ", StringPntr, 3) != 0)
  {
    printf ("False alarm, article line \"of\" misssing in \"%s\".\n",
      LineString);
    return false;
  }
  StringPntr += 2;
  while (*StringPntr == ' ')
    StringPntr++;

  /* Skip over the newsgroup name (no spaces) to the colon. */

  while (*StringPntr != ':' && *StringPntr != ' ' && *StringPntr != 0)
    StringPntr++;

  if (StringPntr[0] != ':' || StringPntr[1] != 0)
  {
    printf ("False alarm, article doesn't end with a colon in \"%s\".\n",
      LineString);
    return false;
  }

  return true;
}



/******************************************************************************
 * Saves the message text to a file in the output directory.  The file name is
 * derived from the message headers.  Returns zero if successful, a negative
 * error code if an error occured.
 */


status_t SaveMessage (BString &MessageText)
{
  int         BodySize;
  char        DateAsString [40];
  time_t      DateInSeconds;
  char        ErrorMessage [B_PATH_NAME_LENGTH + 80];
  int         FirstMark;
  int         HeaderSize;
  BString     HeaderValues [STD_HDR_MAX];
  int         i;
  int         LastMark;
  int         Length;
  char        Letter;
  const char *LineEndPntr;
  const char *LineStartPntr;
  const char *MessagePntr;
  BFile       OutputFile;
  char        OutputName [B_PATH_NAME_LENGTH];
  status_t    ReturnCode = -1;
  BString     SendersName;
  BString     SendersNameCleaned;
  const char *StringPntr;

  /* Remove blank lines from the end of the message (a pet peeve of mine), but
  end the message with a single new line to avoid annoying text editors that
  like to have it. */

  i = MessageText.Length ();
  while (i > 0 && (MessageText[i-1] == '\n' || MessageText[i-1] == '\r'))
    i--;
  MessageText.Truncate (i);
  MessageText.Append ("\r\n");

  /* Scan through the header for known attributes, and store the associated
  text for each one. */

  MessagePntr = MessageText.String();
  StringPntr = MessagePntr;

  while (*StringPntr != 0)
  {
    LineStartPntr = StringPntr;
    LineEndPntr = StringPntr;

    /* Find the end of line (or end of the message). */

    while (*LineEndPntr != '\r' && *LineEndPntr != 0)
      LineEndPntr++;

    if (LineEndPntr == LineStartPntr)
    {
      if (*StringPntr == '\r')
        StringPntr += 2; /* Include the blank line as part of header. */
      break; /* Found a blank line, end of the headers. */
    }

    /* Find the leading colon, which marks the end of the field name.  Lines
    starting with a space or tab character are continuation lines,
    ignore them. */

    if (!isspace (*StringPntr))
    {
      while (StringPntr < LineEndPntr && *StringPntr != ':')
        StringPntr++;
      if (StringPntr >= LineEndPntr)
      {
        printf ("Odd, line with no colon in the header section: \"%.*s\".\n",
          (int) (LineEndPntr - LineStartPntr), LineStartPntr);
      }
      else /* Found a field with a colon. */
      {
        /* Position StringPntr at the first non-blank value in the field. */

        StringPntr++; /* Skip past the colon. */
        while (StringPntr < LineEndPntr && isspace (*StringPntr))
          StringPntr++;
        if (StringPntr < LineEndPntr) /* If value is non-empty. */
        {
          /* See if it is one of the field names we are capturing.  If so, save
          the field value for later. */

          for (i = 0; i < STD_HDR_MAX; i++)
          {
            if (strncasecmp (LineStartPntr, g_StandardHeaders[i],
            strlen (g_StandardHeaders[i])) == 0)
            {
              HeaderValues[i].SetTo (StringPntr, LineEndPntr - StringPntr);
              break;
            }
          }
        }
      }
    }

    /* Advance to the start of the next line. */

    StringPntr = LineEndPntr;
    if (*StringPntr == '\r')
      StringPntr += 2;
  }

  /* StringPntr is positioned at the start of the message body now.  Save the
  sizes of the parts for later. */

  HeaderSize = StringPntr - MessagePntr;
  BodySize = MessageText.Length() - HeaderSize;

  /* Extract the sender's name from the From: field, if any.  Done by removing
  the e-mail address in <> marks, triming off outlying quotes and spaces.  It
  also handles names in round brackets (the old UseNet style). */

  SendersName.SetTo (HeaderValues [STD_HDR_FROM]);

  FirstMark = SendersName.FindFirst ("(");
  if (FirstMark >= 0)
  {
    LastMark = SendersName.FindFirst (")", FirstMark);
    if (LastMark < 0)
      LastMark = SendersName.Length();
    SendersName.Remove (LastMark, SendersName.Length() - LastMark);
    SendersName.Remove (0, FirstMark + 1);
  }

  FirstMark = SendersName.FindFirst ("<");
  if (FirstMark >= 0)
  {
    LastMark = SendersName.FindLast (">");
    if (LastMark < 0)
      LastMark = SendersName.Length();
    else
      LastMark++;
    SendersName.Remove (FirstMark, LastMark - FirstMark);
  }

  /* Trim leading spaces and quote marks. */

  i = 0;
  Length = SendersName.Length();
  while (i < Length)
  {
    Letter = SendersName[i];
    if (!isspace (Letter) && Letter != '\"' && Letter != '\'')
      break;
    i++;
  }
  if (i > 0)
    SendersName.Remove (0, i);

  /* Trim trailing spaces and quote marks. */

  i = SendersName.Length();
  while (i > 0)
  {
    Letter = SendersName[i-1];
    if (!isspace (Letter) && Letter != '\"' && Letter != '\'')
      break;
    i--;
  }
  if (i < SendersName.Length())
    SendersName.Truncate (i);

  if (SendersName.Length() == 0) /* Got a name with only <myname@domain>. */
    SendersName.SetTo (HeaderValues [STD_HDR_FROM]);

  /* Convert the date into the number of seconds since the start of time.  Glad
  I found that parsedate function! */

  DateInSeconds = parsedate (HeaderValues[STD_HDR_DATE].String(), -1);
  if (DateInSeconds == -1)
    DateInSeconds = 0; /* Set it to the earliest time if date isn't known. */

  /* Get the date into a standard text format, for use as part of the file
  name. */

  strcpy (DateAsString, ctime (&DateInSeconds));
  i = strlen (DateAsString) - 1;
  while (i >= 0 && isspace (DateAsString[i]))
    i--; /* Need to remove trailing new line character etc. */
  DateAsString[i+1] = 0;

  /* Make up a hopefully unique file name for the new e-mail file.  If it isn't
  unique, add a digit at the end and increment it until an unused name is
  found. */

  SendersNameCleaned.SetTo (SendersName);
  SendersNameCleaned.ReplaceSet ("/\\<>", "-"); /* Make it a legal name. */

  sprintf (OutputName, "%s%s - %s",
    OutputDirectoryPathName,
    SendersNameCleaned.String(),
    DateAsString);

  for (i = 1; i < 100; i++)
  {
    ReturnCode = OutputFile.SetTo (OutputName,
      B_WRITE_ONLY | B_CREATE_FILE | B_FAIL_IF_EXISTS);

    if (ReturnCode == 0)
      break; /* Successfully created the file. */

    if (ReturnCode != B_FILE_EXISTS)
    {
      sprintf (ErrorMessage, "Unable to create file named \"%s\"", OutputName);
      DisplayErrorMessage (ErrorMessage, ReturnCode);
      return ReturnCode;
    }

    /* Set up the next file name. */

    sprintf (OutputName, "%s%s - %s - %d",
      OutputDirectoryPathName,
      SendersNameCleaned.String(),
      DateAsString,
      i);
  }

  if (i >= 100)
  {
    sprintf (ErrorMessage, "Unable to find a unique file name, last one "
      "tried was \"%s\".\n", OutputName);
    DisplayErrorMessage (ErrorMessage, ReturnCode);
    return ReturnCode;
  }

  if (i != 1)
    printf ("Had to increment output name \"%s\" to avoid similarly "
    "named files.\n", OutputName);

  /* Write the file contents. */

  i = OutputFile.Write (MessageText.String(), MessageText.Length());
  ReturnCode = errno;
  if (i != MessageText.Length())
  {
    sprintf (ErrorMessage, "Only wrote %d bytes of %d to %s",
      i, (int) MessageText.Length(), OutputName);
    DisplayErrorMessage (ErrorMessage, ReturnCode);
    return ReturnCode;
  }

  /* Write the various attributes, without much error checking. */

  /* Save the MIME type first, otherwise the live queries don't pick up the new
  file.  Theoretically it would be better to do it last so that other programs
  don't start reading the message before the other attributes are set. */

  OutputFile.WriteAttr ("BEOS:TYPE", B_MIME_STRING_TYPE, 0,
    "text/x-email", 13);

  OutputFile.WriteAttr ("MAIL:when", B_TIME_TYPE, 0,
    &DateInSeconds, sizeof (DateInSeconds));

  if ((Length = HeaderValues[STD_HDR_MIME_VERSION].Length()) > 0)
    OutputFile.WriteAttr ("MAIL:mime", B_STRING_TYPE, 0,
    HeaderValues[STD_HDR_MIME_VERSION].String(), Length + 1);

  OutputFile.WriteAttr ("MAIL:header_length", B_INT32_TYPE, 0,
    &HeaderSize, sizeof (HeaderSize));

  OutputFile.WriteAttr ("MAIL:content_length", B_INT32_TYPE, 0,
    &BodySize, sizeof (BodySize));

  if ((Length = HeaderValues[STD_HDR_FROM].Length()) > 0)
    OutputFile.WriteAttr ("MAIL:from", B_STRING_TYPE, 0,
    HeaderValues[STD_HDR_FROM].String(), Length + 1);

  if ((Length = SendersName.Length()) > 0)
    OutputFile.WriteAttr ("MAIL:name", B_STRING_TYPE, 0,
    SendersName.String(), Length + 1);

  if ((Length = HeaderValues[STD_HDR_TO].Length()) > 0)
    OutputFile.WriteAttr ("MAIL:to", B_STRING_TYPE, 0,
    HeaderValues[STD_HDR_TO].String(), Length + 1);

  if ((Length = HeaderValues[STD_HDR_REPLY].Length()) > 0)
    OutputFile.WriteAttr ("MAIL:reply", B_STRING_TYPE, 0,
    HeaderValues[STD_HDR_REPLY].String(), Length + 1);

  if ((Length = HeaderValues[STD_HDR_SUBJECT].Length()) > 0)
    OutputFile.WriteAttr ("MAIL:subject", B_STRING_TYPE, 0,
    HeaderValues[STD_HDR_SUBJECT].String(), Length + 1);

  if ((Length = HeaderValues[STD_HDR_CC].Length()) > 0)
    OutputFile.WriteAttr ("MAIL:cc", B_STRING_TYPE, 0,
    HeaderValues[STD_HDR_CC].String(), Length + 1);

  /* Set the priority, there are several sources so set them from least useful
  to most definitive. */

  if ((Length = HeaderValues[STD_HDR_X_PRIORITY].Length()) > 0)
    OutputFile.WriteAttr ("MAIL:priority", B_STRING_TYPE, 0,
    HeaderValues[STD_HDR_X_PRIORITY].String(), Length + 1);

  if ((Length = HeaderValues[STD_HDR_MSMAIL_PRIORITY].Length()) > 0)
    OutputFile.WriteAttr ("MAIL:priority", B_STRING_TYPE, 0,
    HeaderValues[STD_HDR_MSMAIL_PRIORITY].String(), Length + 1);

  if ((Length = HeaderValues[STD_HDR_PRIORITY].Length()) > 0)
    OutputFile.WriteAttr ("MAIL:priority", B_STRING_TYPE, 0,
    HeaderValues[STD_HDR_PRIORITY].String(), Length + 1);

  OutputFile.WriteAttr ("MAIL:status", B_STRING_TYPE, 0,
    "Read", 5); /* So that old messages don't show up as unread. */

  return 0;
}



/******************************************************************************
 * Finally, the main program which drives it all.
 */

int main (int argc, char** argv)
{
  char         ErrorMessage [B_PATH_NAME_LENGTH + 80];
  bool         HaveOldMessage = false;
  int          MessagesDoneCount = 0;
  BString      MessageText;
  BApplication MyApp ("application/x-vnd.agmsmith.MailboxFileToBeMail");
  status_t     ReturnCode = -1;
  char        *StringPntr;
  char         TempString [102400];

  if (argc <= 1 || argc >= 4)
  {
    printf ("%s is a utility for converting Pine e-mail\n",
      argv[0]);
    printf ("files to BeOS e-mail files with attributes.  It could well\n");
    printf ("work with other Unix style mailbox files.  Each message in\n");
    printf ("the input mailbox is converted into a separate file.  You\n");
    printf ("can optionally specify a directory to put all the output\n");
    printf ("files in.  Usage:\n\n");
    printf ("MailboxFileToBeMail InputFile [OutputDirectory]\n\n");
    printf ("Public domain, by Alexander G. M. Smith.\n");
    printf ("$Header: /boot/home/agmsmith/Programming/MailboxFileToBeMail/RCS/MailboxToBeMail.cpp,v 1.6 2002/01/14 20:23:21 agmsmith Exp $\n");
    return -10;
  }

  /* Try to open the input file. */

  strncpy (InputPathName, argv[1], sizeof (InputPathName) - 1);
  InputFile = fopen (InputPathName, "rb");
  if (InputFile == NULL)
  {
    ReturnCode = errno;
    sprintf (ErrorMessage, "Unable to open file \"%s\" for reading",
      InputPathName);
    DisplayErrorMessage (ErrorMessage, ReturnCode);
    goto ErrorExit;
  }

  /* Try to make the output directory. */

  if (argc >= 3)
  {
    strncpy (OutputDirectoryPathName, argv[2],
      sizeof (OutputDirectoryPathName) - 2
      /* Leave space for adding trailing slash and NUL byte */);

    /* Remove trailing '/' characters. */

    StringPntr =
      OutputDirectoryPathName + (strlen (OutputDirectoryPathName) - 1);
    while (StringPntr >= OutputDirectoryPathName)
    {
      if (*StringPntr != '/')
        break;
      StringPntr--;
    }
    *(++StringPntr) = 0;

    if (StringPntr - OutputDirectoryPathName > 0)
    {
      if (mkdir (OutputDirectoryPathName, 0777))
      {
        ReturnCode = errno;
        if (ReturnCode != B_FILE_EXISTS)
        {
          sprintf (ErrorMessage, "Unable to make output directory \"%s\"",
            OutputDirectoryPathName);
          DisplayErrorMessage (ErrorMessage, ReturnCode);
          goto ErrorExit;
        }
      }
      strcat (OutputDirectoryPathName, "/");
    }
  }

  printf ("Input file: \"%s\", Output directory prefix: \"%s\".  Working",
    InputPathName, OutputDirectoryPathName);

  /* Extract a text message from the mail file.  It starts with a line that
  says "From blah Day MM NN XX:XX:XX YYYY TZONE".  Blah is an e-mail address
  you can ignore (just treat it as a word separated by spaces).  Day is a 3
  letter day of the week.  MM is a 3 letter month name.  NN is the two digit
  day of the week, has a leading space if the day is less than 10.  XX:XX:XX is
  the time, the X's are digits.  YYYY is the four digit year.  TZONE is the
  optional time zone info, a plus or minus sign and 4 digits. */

  while (!feof (InputFile))
  {
    /* First read in one line of text. */

    if (!fgets (TempString, sizeof (TempString), InputFile))
    {
      ReturnCode = errno;
      if (ferror (InputFile))
      {
        sprintf (ErrorMessage,
          "Error while reading from \"%s\"", InputPathName);
        DisplayErrorMessage (ErrorMessage, ReturnCode);
        goto ErrorExit;
      }
      break; /* No error, just end of file. */
    }

    /* Remove any trailing control characters (line feed usually, or CRLF).
    Might also nuke trailing tabs too.  Doesn't usually matter.  The main thing
    is to allow input files with both LF and CRLF endings (and even CR endings
    if you come from the Macintosh world). */

    StringPntr = TempString + strlen (TempString) - 1;
    while (StringPntr >= TempString && *StringPntr < 32)
      StringPntr--;
    *(++StringPntr) = 0;

    /* See if this is the start of a new message. */

    if (IsStartOfUsenetArticle (TempString) ||
    IsStartOfMailMessage (TempString))
    {
      if (HaveOldMessage)
      {
        if ((ReturnCode = SaveMessage (MessageText)) != 0)
          goto ErrorExit;
        putchar ('.');
        fflush (stdout);
        MessagesDoneCount++;
      }
      MessageText.SetTo ("");
      HaveOldMessage = true;
    }

    /* Append the line to the current message text.  This includes the
    detection line, so that any info in it isn't lost. */

    if (MessageText.Length () > 0)
      MessageText.Append ("\r\n"); /* Yes, BeMail expects CR/LF line ends. */
    MessageText.Append (TempString);
  }

  /* Flush out the last message. */

  if (HaveOldMessage)
  {
    if ((ReturnCode = SaveMessage (MessageText)) != 0)
      goto ErrorExit;
    putchar ('.');
    MessagesDoneCount++;
  }
  printf ("  Did %d messages.\n", MessagesDoneCount);

  ReturnCode = 0;

ErrorExit:
  if (InputFile != NULL)
    fclose (InputFile);
  return ReturnCode;
}
